import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))

from orso.dataframe import DataFrame
from tests import cities


def test_to_polars():
    import polars

    odf = DataFrame(cities.values)
    pdf = odf.polars()

    assert len(pdf) == 20
    assert "name" in pdf.columns
    assert type(pdf) == polars.DataFrame

    pdf = odf.polars(size=4)
    assert len(pdf) == 4
    assert "name" in pdf.columns
    assert type(pdf) == polars.DataFrame


if __name__ == "__main__":  # prgama: nocover
    from tests import run_tests

    run_tests()



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))

import pytest
from orso import Row
from orso.exceptions import DataError
from orso.row import extract_columns, MAXIMUM_RECORD_SIZE


def test_row_from_dict():
    row_factory = Row.create_class(
        (
            "col1",
            "col2",
        )
    )
    r = row_factory([1, "a"])
    assert r[0] == 1 and r[1] == "a"


def test_row_as_dict():
    row_factory = Row.create_class(
        (
            "col1",
            "col2",
        )
    )
    r = row_factory([1, "a"])
    assert r.as_dict == {"col1": 1, "col2": "a"}


def test_row_as_map():
    row_factory = Row.create_class(
        (
            "col1",
            "col2",
        )
    )
    r = row_factory([1, "a"])
    assert r.as_map == (("col1", 1), ("col2", "a")), r.as_map


def test_row_from_bytes():
    row_factory = Row.create_class(
        (
            "col1",
            "col2",
        )
    )
    original_row = row_factory([1, "a"])
    serialized_row = original_row.as_bytes
    deserialized_row = Row.from_bytes(serialized_row)
    assert original_row == deserialized_row


def test_row_to_bytes():
    row_factory = Row.create_class(
        (
            "col1",
            "col2",
        )
    )
    r = row_factory([1, "a"])
    byte_result = r.as_bytes
    assert isinstance(byte_result, bytes)


def test_row_to_bytes_size_limit():
    big_row = Row(["a" * MAXIMUM_RECORD_SIZE])
    with pytest.raises(DataError):
        big_row.as_bytes


def test_row_to_json():
    row_factory = Row.create_class(
        (
            "col1",
            "col2",
        )
    )
    r = row_factory([1, "a"])
    json_result = r.as_json
    assert json_result == b'{"col1":1,"col2":"a"}'


def test_extract_columns():
    table = [
        Row({"col1": 1, "col2": "a"}),
        Row({"col1": 2, "col2": "b"}),
        Row({"col1": 3, "col2": "c"}),
    ]
    result = extract_columns(table, [0, 1])
    assert result == ([1, 2, 3], ["a", "b", "c"])


if __name__ == "__main__":  # prgama: nocover
    from tests import run_tests

    run_tests()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from dataclasses import _MISSING_TYPE
from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import field
from dataclasses import fields
from decimal import Decimal
from enum import Enum
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import MutableMapping
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union
from warnings import warn

import numpy
from data_expectations import Expectation

from orso.exceptions import ColumnDefinitionError
from orso.exceptions import DataValidationError
from orso.tools import arrow_type_map
from orso.tools import random_string
from orso.types import ORSO_TO_PYTHON_MAP
from orso.types import PYTHON_TO_ORSO_MAP
from orso.types import OrsoTypes

_MISSING_VALUE: str = str()


class SchemaExpectation(Expectation):
    column = _MISSING_VALUE

    @classmethod
    def load(cls: Type["Expectation"], serialized: Union[Dict[str, Any], str]) -> "Expectation":
        """
        Loads a serialized Expectation and returns it as an instance.

        Parameters:
            serialized: Serialized Expectation as a dictionary or JSON string.

        Returns:
            An Expectation instance populated with the serialized data.
        """
        import json
        from copy import deepcopy

        if isinstance(serialized, str):
            serialized = dict(json.loads(serialized))
        serialized_copy: dict = deepcopy(serialized)
        if "expectation" not in serialized_copy:
            raise ValueError("Missing 'expectation' key in Expectation.")
        expectation = serialized_copy.pop("expectation")
        column = serialized_copy.pop("column", _MISSING_VALUE)
        ignore_nulls = serialized_copy.pop("ignore_nulls", True)
        config = serialized_copy
        return Expectation(
            expectation=expectation, column=column, ignore_nulls=ignore_nulls, config=config
        )


@dataclass(init=False)
class FlatColumn:
    """
    This is a standard column type.
    Unlike the other column types, we don't store the values for this Column
    here, we read them from the underlying data structure (orso/pyarrow/velox).
    """

    name: str
    type: OrsoTypes
    description: Optional[str] = None
    aliases: Optional[List[str]] = field(default_factory=list)  # type: ignore
    nullable: bool = True
    expectations: Optional[Expectation] = field(default_factory=list)
    identity: str = field(default_factory=random_string)
    precision: Optional[int] = None
    scale: Optional[int] = None

    def __init__(self, **kwargs):
        attributes = {f.name: f for f in fields(self.__class__)}
        for attribute in attributes:
            if attribute in kwargs:
                value = kwargs[attribute]
                # Special handling for 'expectations'
                if attribute == "expectations" and value:
                    value = [
                        v
                        if isinstance(v, Expectation)
                        else SchemaExpectation.load(v).update({"column": kwargs["name"]})
                        if v.get("column", _MISSING_VALUE) == _MISSING_VALUE
                        else v
                        for v in value
                    ]

                setattr(self, attribute, value)
            elif not isinstance(attributes[attribute].default, _MISSING_TYPE):
                setattr(self, attribute, attributes[attribute].default)
            elif not isinstance(attributes[attribute].default_factory, _MISSING_TYPE):
                setattr(self, attribute, attributes[attribute].default_factory())  # type:ignore
            else:
                raise ColumnDefinitionError(attribute)

        # map literals to OrsoTypes
        if self.type.__class__ is not OrsoTypes:
            type_name = str(self.type).upper()
            if type_name in OrsoTypes.__members__.keys():
                self.type = OrsoTypes[type_name]
            elif type_name == "LIST":
                warn("Column type LIST will be deprecated in a future version, use ARRAY instead.")
                self.type = OrsoTypes.ARRAY
            elif type_name == "NUMERIC":
                warn(
                    "Column type NUMERIC will be deprecated in a future version, use DECIMAL, DOUBLE or INTEGER instead. Mapped to DOUBLE, this may not be compatible with all values NUMERIC was compatible with."
                )
                self.type = OrsoTypes.DOUBLE
            elif type_name == "STRING":
                raise ValueError(
                    f"Unknown column type '{self.type}' for column '{self.name}'. Did you mean 'VARCHAR'?"
                )
            elif self.type != 0:
                raise ValueError(f"Unknown column type '{self.type}' for column '{self.name}'.")

    def __str__(self):
        return self.identity

    def materialize(self):
        raise TypeError("Cannot materialize FlatColumns")

    @classmethod
    def from_arrow(cls, arrow_field) -> "FlatColumn":
        """
        Converts a PyArrow field to an Orso FlatColumn object.

        Parameters:
            arrow_field: Field
                PyArrow Field object to be converted.

        Returns:
            FlatColumn: A FlatColumn object containing the converted information.
        """
        # Fetch the native type mapping from Arrow to Python native types
        native_type = arrow_type_map(arrow_field.type)
        # Initialize variables to hold optional decimal properties
        scale: Optional[int] = None
        precision: Optional[int] = None
        # Check if the type is Decimal and populate scale and precision
        if isinstance(native_type, Decimal):
            field_type = OrsoTypes.DECIMAL
            scale = native_type.scale  # type:ignore
            precision = native_type.precision  # type:ignore
        else:
            # Fall back to the generic mapping
            field_type = PYTHON_TO_ORSO_MAP.get(native_type, None)
            if field_type is None:
                raise ValueError(f"Unsupported type: {native_type}")

        return FlatColumn(
            name=str(arrow_field.name),
            type=field_type,
            nullable=arrow_field.nullable,
            scale=scale,
            precision=precision,
        )

    def to_flatcolumn(self) -> "FlatColumn":
        """
        Convert any column type to a FlatColumn (e.g. after a FunctionColumn has been evaluated)
        """
        return FlatColumn(
            name=str(self.name),
            description=self.description,
            aliases=self.aliases,
            identity=self.identity,
            type=self.type,
            nullable=self.nullable,
            scale=self.scale,
            precision=self.precision,
        )


@dataclass(init=False)
class FunctionColumn(FlatColumn):
    """
    This is a virtual column, it's nominally a column where the value is
    derived from a function.
    """

    binding: Optional[Callable] = lambda: None
    configuration: Tuple = field(default_factory=tuple)
    length: int = 1

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def materialize(self):
        """
        Turn this virtual column into a list
        """
        value = self.binding(*self.configuration)
        return numpy.array([value] * self.length)


@dataclass(init=False)
class ConstantColumn(FlatColumn):
    """
    Rather than pass around columns of constant values, where we can we should
    replace them with this column type.

    note we don't implement anything here which deals with doing operations on
    two constant columns; whilst that would be a good optimization, the better
    way to do this is in the query optimizer, do operations on two constants
    while we're still working with a query plan.
    """

    length: int = 1
    value: Any = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def materialize(self):
        """
        Turn this virtual column into a list
        """
        return numpy.array([self.value] * self.length)


@dataclass(init=False)
class DictionaryColumn(FlatColumn):
    """
    If we know a column has a small amount of unique values AND is a large column
    AND we're going to perform an operation on the values, we should dictionary
    encode the column. This allows us to operate once on each unique value in the
    column, rather than each value in the column individually. At the cost of
    constructing and materializing the dictionary encoding.
    """

    values: List[Any] = field(default_factory=list)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        values = numpy.asarray(self.values)
        self.dictionary, self.encoding = numpy.unique(values, return_inverse=True)

    def materialize(self):
        """
        Turn this virtual column into a list
        """
        return self.dictionary[self.encoding]


@dataclass
class RelationSchema:
    name: str
    aliases: List[str] = field(default_factory=list)
    columns: List[FlatColumn] = field(default_factory=list)

    def __iter__(self):
        """Return an iterator over column names."""
        return iter([col.name for col in self.columns])

    def __add__(self, other: "RelationSchema") -> "RelationSchema":
        """
        When we add schemas together, we combine the set of columns.

        Parameters:
            other: RelationSchema
                The other schema to be added.

        Returns:
            RelationSchema: A new RelationSchema containing the combined columns.
        """
        new_schema = RelationSchema(name=self.name, aliases=self.aliases, columns=[])

        # Create a new list to hold the merged columns
        new_columns = self.columns[:]

        # Keep track of the seen identities - preload with the current set
        seen_identities = [col.identity for col in self.columns]

        for column in other.columns:
            if column.identity not in seen_identities:
                seen_identities.append(column.identity)
                new_columns.append(column)

        # Assign the new list of columns to the new schema
        new_schema.columns = new_columns

        return new_schema

    @property
    def num_columns(self) -> int:
        """Returns the number of columns in the schema."""
        return len(self.columns)

    def find_column(self, column_name: str) -> Optional[FlatColumn]:
        """
        Find a column by name or alias.

        Parameters:
            column_name: str
                Name or alias of the column.

        Returns:
            Optional[FlatColumn]: The FlatColumn object, if found. None otherwise.
        """
        for column in self.columns:
            if column.name == column_name:
                return column
            if column_name in column.aliases:
                return column
        return None

    def all_column_names(self) -> List[str]:
        """
        Return all the names and aliases for columns in this relation.

        Returns:
            List[str]: List of all names and aliases.
        """

        def _inner():
            for column in self.columns:
                yield column.name
                yield from column.aliases

        return list(_inner())

    @property
    def column_names(self) -> List[str]:
        """Return the names for columns in this relation."""
        return [column.name for column in self.columns]

    def column(self, i: Union[int, str]) -> Optional[FlatColumn]:
        """
        Get column by name or index.

        Parameters:
            i: Union[int, str]
                Index or name of the column.

        Returns:
            Optional[FlatColumn]: The FlatColumn object, if found. None otherwise.
        """
        if isinstance(i, int):
            return self.columns[i]
        else:
            return self.find_column(i)

    def pop_column(self, column_name: str) -> Optional[FlatColumn]:
        """
        Remove a column by its name and return it.

        Parameters:
            column_name: str
                Name of the column to be removed.

        Returns:
            Optional[FlatColumn]: The removed column if found, otherwise None.
        """
        for idx, column in enumerate(self.columns):
            if column.name == column_name:
                return self.columns.pop(idx)
        return None

    def to_dict(self) -> Dict:
        """
        Convert the Schema to a dictionary.

        Returns:
            Dict: A dictionary representation of the schema.
        """

        def _converter(obj) -> Dict:
            """Handle enum serialization."""
            return {key: value.value if isinstance(value, Enum) else value for key, value in obj}

        return asdict(self, dict_factory=_converter)

    @classmethod
    def from_dict(cls, dic: Dict) -> "RelationSchema":
        """
        Create a Schema from a dictionary.

        Parameters:
            dic: Dict
                A dictionary to convert into a RelationSchema.

        Returns:
            RelationSchema: A new RelationSchema object.
        """
        schema = RelationSchema(name=dic["name"], aliases=dic.get("aliases", []))
        for column in dic["columns"]:
            schema.columns.append(FlatColumn(**column))
        return schema

    def validate(self, data: MutableMapping) -> bool:
        """
        Perform schema validation against a dictionary-formatted record.

        Parameters:
            data: MutableMapping
                A dictionary containing the data to validate against the schema.

        Returns:
            bool: True if the data is valid according to the schema.

        Raises:
            TypeError: If data is not dictionary-like.
            DataValidationError: If data validation fails.
        """
        if not isinstance(data, MutableMapping):
            raise TypeError("Cannot validate non Dictionary-type value")

        for column in self.columns:
            if column.name not in data:
                raise DataValidationError(column=column, value=None, error="Column Missing")

            value = data[column.name]

            if value is None:
                if not column.nullable:
                    raise DataValidationError(
                        column=column, value=value, error="None not acceptable"
                    )
            else:
                if not isinstance(value, ORSO_TO_PYTHON_MAP.get(column.type)):
                    raise DataValidationError(column=column, value=value, error="Incorrect Type")

        return True



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))

from orso.filters.cuckoo_filter import CuckooFilter

ITERATIONS: int = 50000


def random_string(width: int = 16):
    import random

    num_chars = ((width + 1) >> 1) << 3  # Convert length to number of bits
    rand_bytes = random.getrandbits(num_chars)  # Generate random bytes
    rand_hex = hex(rand_bytes)[
        2 : width + 2
    ]  # Convert to hex string and truncate to desired length
    return rand_hex


def test_cuckoo_filter():
    # first we populate the BloomFilter
    tokens = (random_string(48) for i in range(ITERATIONS))
    bf = CuckooFilter()
    for token in tokens:
        bf.add(token)

    # then we test. 100% shouldn't match (we use different string lengths)
    # but we're a probabilistic filter so expect some false positives
    tokens = (random_string(32) for i in range(ITERATIONS))
    collisions = 0
    for token in tokens:
        if token in bf:
            collisions += 1

    # should be about 0.006
    assert (ITERATIONS * 0.001) < collisions < (ITERATIONS * 0.01), collisions / ITERATIONS


def test_cuckoo_contains():
    bf = CuckooFilter()
    bf.add("test")
    assert "test" in bf
    assert "nonexistent" not in bf


if __name__ == "__main__":  # prgama: nocover
    from tests import run_tests

    run_tests()



"""
from: https://stackoverflow.com/a/35804945

#nodoc - don't add to the documentation wiki
"""

import atexit
import logging
from typing import Dict

import orjson

logging_seen_warnings: Dict[int, int] = {}


def report_suppressions(message):
    from .. import logging as ml

    record = logging_seen_warnings.get(hash(message))

    if record:
        ml.get_logger().warning(
            f'The following message was suppressed {record} time(s) - "{message}"'
        )


def add_logging_level(level_name, level_num, method_name=None):
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `level_name` becomes an attribute of the `logging` module with the value
    `level_num`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `method_name` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example:
        addLoggingLevel('TRACE', logging.DEBUG - 5)
        logging.getLogger(__name__).setLevel("TRACE")
        logging.getLogger(__name__).trace('that worked')
        logging.trace('so did this')
        logging.TRACE

    """
    if not method_name:
        method_name = level_name.lower()

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def log_for_level(self, message, *args, **kwargs):
        # if we've added the level,it doesn't format the message as JSON
        if isinstance(message, dict):
            try:
                message = orjson.dumps(message)
            except:
                message = str(message)
        # not just the json decoder outputs in bytes, make it a string
        if isinstance(message, bytes):
            message = message.decode()
        if self.isEnabledFor(level_num):
            # supress duplicate warnings
            if level_num == 30:  # warnings
                hashed = hash(message)
                if hashed in logging_seen_warnings:
                    logging_seen_warnings[hashed] += 1
                    return
                logging_seen_warnings[hashed] = 0
                atexit.register(report_suppressions, message)

            self._log(level_num, message, args, **kwargs)

    def log_to_root(message, *args, **kwargs):
        logging.log(level_num, message, *args, **kwargs)

    logging.addLevelName(level_num, level_name)
    setattr(logging, level_name, level_num)
    setattr(logging.getLoggerClass(), method_name, log_for_level)
    setattr(logging, method_name, log_to_root)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ---------------------------------------------------------------------------------
# Our record when converted to bytes looks like this
#
#       ┌───────────┬─────────────┬──────────────┐
#       │   flags   │   row_size  │  row values  │
#       └───────────┴─────────────┴──────────────┘
#
# 'flags' currently has a a version string and a single flag of deleted.
#
# Row object looks and acts like a Tuple where possible, but has additional features
# such as as_dict() to render as a dictionary.

import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

import numpy
import orjson
from ormsgpack import OPT_SERIALIZE_NUMPY
from ormsgpack import packb

from orso.compiled import from_bytes_cython
from orso.exceptions import DataError
from orso.schema import RelationSchema

HEADER_SIZE: int = 6
HEADER_PREFIX: bytes = b"\x10\x00"
MAXIMUM_RECORD_SIZE: int = 8 * 1024 * 1024


def extract_columns(table: List[Dict[str, Any]], columns: List[str]) -> Tuple[List[Any], ...]:
    """
    Extracts specified columns from a table.

    Parameters:
        table: List of dictionaries representing rows in a table.
        columns: List of column names to extract.
    Returns:
        A tuple of lists, each containing values of a particular column.
    """
    n_rows = len(table)
    result = [None] * len(columns)
    for i in range(len(columns)):
        result[i] = [None] * n_rows

    for i, row in enumerate(table):
        for j, column in enumerate(columns):
            result[j][i] = row[column]

    return tuple(result)


class Row(tuple):
    __slots__ = ()
    _fields: Tuple[str, ...] = None
    _cached_map: Tuple[Tuple[str, Any]] = None

    def __new__(cls, data: Union[Dict[str, Any], Tuple[Any, ...]]):
        """
        Creates a new Row instance.

        Parameters:
            data: Either a dictionary or a tuple.
        Returns:
            A new Row instance.
        """
        if isinstance(data, dict):
            data = data.values()
        instance = super().__new__(cls, data)
        return instance

    @property
    def as_map(self) -> Tuple[Tuple[str, Any], ...]:
        """
        Returns the Row as a tuple of key-value pair tuples (a 'map').

        Returns:
            A tuple of key-value pair tuples.
        """
        if self._cached_map is None:
            self._cached_map = tuple(zip(self._fields, self))
        return self._cached_map

    @property
    def as_dict(self) -> Dict[str, Any]:
        """
        Returns the Row as a dictionary.

        Returns:
            A dictionary representation of the Row.
        """
        return dict(self.as_map)

    @property
    def values(self) -> Tuple[Any, ...]:
        return tuple(self)

    def keys(self) -> Tuple[str, ...]:
        return self._fields

    @classmethod
    def from_bytes(cls, data: bytes) -> "Row":
        """
        Creates a Row instance from bytes.

        Parameters:
            data: The byte representation of a Row.
        Returns:
            A new Row instance.
        """
        return cls(from_bytes_cython(data))

    @property
    def as_bytes(self) -> bytes:
        """
        Converts the Row to bytes.

        Returns:
            The byte representation of the Row.
        """

        def serialize(value):
            if isinstance(value, numpy.datetime64):
                if numpy.isnat(value):
                    return None
                return ("__datetime__", value.astype("datetime64[s]").astype("int"))
            if isinstance(value, datetime.datetime):
                return ("__datetime__", value.timestamp())
            if isinstance(value, numpy.ndarray):
                return list(value)
            return str(value)

        record_bytes = packb(tuple(self), option=OPT_SERIALIZE_NUMPY, default=serialize)
        record_size = len(record_bytes)
        if record_size > MAXIMUM_RECORD_SIZE:
            raise DataError("Record length cannot exceed 8Mb")

        return HEADER_PREFIX + record_size.to_bytes(4, "big") + record_bytes

    @property
    def as_json(self) -> bytes:
        """
        Converts the Row to JSON bytes.

        Returns:
            The JSON byte representation of the Row.
        """
        return orjson.dumps(self.as_dict, default=str)

    @classmethod
    def create_class(cls, schema: Union[RelationSchema, Tuple[str, ...], List[str]]) -> type:
        """
        Creates a subclass of Row based on the schema provided.

        Parameters:
            schema: Either a RelationSchema or a list/tuple of field names.
        Returns:
            A new Row subclass.
        """
        if isinstance(schema, RelationSchema):
            fields = tuple(c.name for c in schema.columns)
        elif isinstance(schema, (list, tuple)):
            fields = tuple(schema)
        else:
            raise ValueError("Row requires either a list of field names or a RelationSchema")

        return type("RowFactory", (Row,), {"_fields": fields})



import hashlib
import json
import logging
import os
import re

from orso.display import colorizer

try:
    # added 3.9
    from functools import cache
except ImportError:
    from functools import lru_cache

    cache = lru_cache(1)


# if we find a key which matches these strings, we hash the contents
KEYS_TO_SANITIZE = [
    r"password$",
    r"pwd$",
    r".*_secret$",
    r".*_key$",
    r"_token$",
    r"credentials",
]
COLOR_EXCHANGES = {
    " ALERT    ": "\001BOLD_REDm ALERT    \001OFFm",
    " ERROR    ": "\001REDm ERROR    \001OFFm",
    " DEBUG    ": "\001GREENm DEBUG    \001OFFm",
    " AUDIT    ": "\001YELLOWm AUDIT    \001OFFm",
    " WARNING  ": "\001BOLD_REDm WARNING  \001OFFm",
    " INFO     ": "\001BOLD_WHITEm INFO     \001OFFm",
}


class LogFormatter(logging.Formatter):
    def __init__(self, orig_formatter, suppress_color: bool = False):
        """
        Remove sensitive data from records before saving to external logs. Note that
        the value is hashed using (SHA256) and only the first 8 characters of the
        hex-encoded hash are presented. This information allows values to be traced
        without disclosing the actual value.

        The Sanitizer can only sanitize dictionaries, it doesn't sanitize strings,
        which could contain sensitive information We use the message id as a salt to
        further protect sensitive information.

        Based On: https://github.com/joocer/cronicl/blob/main/cronicl/utils/sanitizer.py
        """
        self.orig_formatter = orig_formatter
        self.suppress_color = suppress_color

    def format(self, record):
        try:
            msg = self.orig_formatter.format(record)
        except:
            msg = record
        msg = self.sanitize_record(msg)
        if "://" in msg:
            msg = re.sub(r":\/\/(.*?)\@", r"://\001BOLD_PURLEm<redacted>\001OFFm", msg)
        return msg

    @cache
    def _can_colorize(self):
        if self.suppress_color:
            return False

        colorterm = os.environ.get("COLORTERM", "").lower()
        term = os.environ.get("TERM", "").lower()
        return "yes" in colorterm or "true" in colorterm or "256" in term

    def color_code(self, record):
        if self._can_colorize():
            for k, v in COLOR_EXCHANGES.items():
                if k in record:
                    return record.replace(k, v)
        return record

    def __getattr__(self, attr):
        return getattr(self.orig_formatter, attr)

    def hash_it(self, value_to_hash):
        return hashlib.sha256(value_to_hash.encode()).hexdigest()[:8]

    def clean_record(self, dirty_record, colorize: bool = True):
        if colorize:
            KEY = "\001KEYm"
            OFF = "\001OFFm"
            PURPLE = "\001PURPLEm"
            YELLOW = "\001YELLOWm"
            VALUE = "\001VALUEm"
        else:
            KEY = ""
            OFF = ""
            PURPLE = ""
            YELLOW = ""
            VALUE = ""

        clean_record = {}
        for key, value in dirty_record.items():
            if isinstance(value, dict):
                value = self.clean_record(value, colorize)
            elif any(
                [
                    True
                    for expression in KEYS_TO_SANITIZE
                    if re.match(expression, key, re.IGNORECASE)
                ]
            ):
                clean_record[KEY + key + OFF] = (
                    PURPLE + "<redacted:" + self.hash_it(str(value)) + ">" + OFF
                )
            else:
                value = re.sub(r"`([^`]*)`", "`" + YELLOW + "\1" + VALUE + r"`", f"{value}")
                value = re.sub(r"'([^']*)'", "'" + YELLOW + "\1" + VALUE + r"'", f"{value}")
                clean_record[str(KEY) + key + OFF] = VALUE + f"{value}" + OFF
        return clean_record

    def sanitize_record(self, record):
        record = self.color_code(record)
        parts = record.split("|")
        json_part = parts.pop()

        try:
            dirty_record = json.loads(json_part.encode("UTF8"))
            clean_record = self.clean_record(dirty_record)
            parts.append(" " + json.dumps(clean_record))

        except ValueError:
            json_part = re.sub(r"`([^`]*)`", r"`\001YELLOWm\1\001OFFm`", f"{json_part}")
            json_part = re.sub(r"'([^']*)'", r"'\001YELLOWm\1\001OFFm'", f"{json_part}")
            json_part = re.sub(r'"([^"]*)"', r"'\001YELLOWm\1\001OFFm'", f"{json_part}")
            parts.append(" " + json_part.strip() + " *")

        record = "|".join(parts)
        return colorizer(record, self._can_colorize())



import os
import sys
import time
from unittest.mock import Mock

sys.path.insert(1, os.path.join(sys.path[0], ".."))

from orso.tools import single_item_cache


def test_single_item_cache():
    mock_func = Mock(side_effect=lambda *args, **kwargs: args[0] if args else kwargs.get("arg"))
    cached_func = single_item_cache(mock_func)

    # Test 1: Basic functionality
    assert cached_func("test") == "test"
    assert mock_func.call_count == 1

    # Test 2: Cache hit
    assert cached_func("test") == "test"
    assert mock_func.call_count == 1

    # Test 3: Cache miss, new argument
    assert cached_func("new test") == "new test"
    assert mock_func.call_count == 2

    # Test 4: Cache hit, same argument as before
    assert cached_func("new test") == "new test"
    assert mock_func.call_count == 2

    # Test 5: Cache miss, old argument
    assert cached_func("test") == "test"
    assert mock_func.call_count == 3, mock_func.call_count

    # Test 6: Complex arguments
    assert cached_func(("test",)) == ("test",)
    assert mock_func.call_count == 4

    # Test 7: Cache hit, complex arguments
    assert cached_func(("test",)) == ("test",)
    assert mock_func.call_count == 4

    # Test 8: Keyword arguments
    assert cached_func(arg="test") == "test"
    assert mock_func.call_count == 5

    # Test 9: Cache hit, keyword arguments
    assert cached_func(arg="test") == "test"
    assert mock_func.call_count == 5


def test_single_item_cache_with_same_arguments():
    counter = [0]

    @single_item_cache()
    def foo(x):
        counter[0] += 1
        return x * 2

    assert foo(1) == 2
    assert counter[0] == 1

    assert foo(1) == 2
    assert counter[0] == 1


def test_single_item_cache_with_different_arguments():
    counter = [0]

    @single_item_cache()
    def foo(x):
        counter[0] += 1
        return x * 2

    assert foo(1) == 2
    assert counter[0] == 1

    assert foo(2) == 4
    assert counter[0] == 2


def test_single_item_cache_with_same_then_different_then_same_arguments():
    counter = [0]

    @single_item_cache()
    def foo(x):
        counter[0] += 1
        return x * 2

    assert foo(1) == 2
    assert counter[0] == 1

    assert foo(2) == 4
    assert counter[0] == 2

    assert foo(1) == 2
    assert counter[0] == 3


def test_single_item_cache_with_expiration():
    counter = [0]

    @single_item_cache(valid_for_seconds=1)
    def foo(x):
        counter[0] += 1
        return x * 2

    assert foo(1) == 2
    assert counter[0] == 1

    assert foo(1) == 2
    assert counter[0] == 1

    time.sleep(1.1)

    assert foo(1) == 2
    assert counter[0] == 2


def test_single_item_cache_with_nested_arguments():
    mock_func = Mock(side_effect=lambda *args: args[0][0])
    cached_func = single_item_cache(mock_func)

    # Test 10: Cache miss, nested arguments
    assert cached_func(("nested",)) == "nested"
    assert mock_func.call_count == 1

    # Test 11: Cache hit, nested arguments
    assert cached_func(("nested",)) == "nested"
    assert mock_func.call_count == 1


def test_single_item_cache_with_function_calls():
    mock_func = Mock(side_effect=lambda x: x)
    cached_func1 = single_item_cache(mock_func)
    cached_func2 = single_item_cache(mock_func)

    # Test 12: Cache miss, function call within function call
    assert cached_func1(cached_func2("nesting")) == "nesting"
    assert mock_func.call_count == 2

    # Test 13: Cache hit, function call within function call
    assert cached_func1(cached_func2("nesting")) == "nesting"
    assert mock_func.call_count == 2  # Should not increment


def test_single_item_cache_with_mixed_args_kwargs():
    mock_func = Mock(side_effect=lambda *args, **kwargs: args[0] if args else kwargs.get("arg"))
    cached_func = single_item_cache(mock_func)

    # Test 14: Cache miss, mixed args and kwargs
    assert cached_func("test", arg="test") == "test"
    assert mock_func.call_count == 1

    # Test 15: Cache hit, mixed args and kwargs
    assert cached_func("test", arg="test") == "test"
    assert mock_func.call_count == 1  # Should not increment

    # Test 16: Cache miss, different kwargs
    assert cached_func("test", arg="different") == "test"
    assert mock_func.call_count == 2


def test_single_item_cache_invalidates_on_none():
    mock_func = Mock(side_effect=lambda *args: args[0])
    cached_func = single_item_cache(mock_func)

    # Test 17: Cache miss, argument is None
    assert cached_func(None) == None
    assert mock_func.call_count == 1

    # Test 18: Cache miss, different argument after None
    assert cached_func("test") == "test"
    assert mock_func.call_count == 2


if __name__ == "__main__":  # pragma: nocover
    from tests import run_tests

    run_tests()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import decimal
from typing import Union

# Background		#282a36	40 42 54	231° 15% 18%
# Current Line		#44475a	68 71 90	232° 14% 31%
# Foreground		#f8f8f2	248 248 242	60° 30% 96%
# Red		        #ff5555	255 85 85	0° 100% 67%
# Yellow		    #f1fa8c	241 250 140	65° 92% 76%

COLORS = {
    "\001OFFm": "\033[0m",  # Text Reset
    # Opteryx named colors
    "\001PUNCm": "\033[38;5;102m",
    "\001CRLFm": "\033[38;2;98;114;164m",
    "\001HEADm": "\033[1m",
    "\001VARCHARm": "\033[38;2;255;171;82m",  # orange
    "\001CONSTm": "\033[38;2;139;233;253m\033[3m",  # cyan, italic
    "\001NULLm": "\033[38;2;98;114;164m\033[3m",  # grey, italic
    "\001TYPEm": "\033[38;2;98;114;164m",  # grey,
    "\001VALUEm": "\033[38;2;139;233;253m",  # cyan
    "\001FLOATm": "\033[38;2;255;121;198m",  # pink
    "\001INTEGERm": "\033[38;2;189;147;249m",  # purple
    "\001DATEm": "\033[38;2;80;250;123m",  # green
    "\001TIMESTAMPm": "\033[38;2;80;250;123m",  # green
    "\001TIMEm": "\033[38;2;26;185;67m",  # non-std green
    "\001KEYm": "\033[38;2;189;147;249m",  # purple
    "\001BLOBm": "\033[38;2;241;250;140m",  # yellow
    "\001INTERVALm": "\033[38;2;255;85;85m",  # pink
    # an orange color - 222
    # a red color = 209
    # Regular Colors
    "\001BLACKm": "\033[0;30m",  # Black
    "\001REDm": "\033[38;5;203m",  # Red
    "\001GREENm": "\033[38;2;80;250;123m",  # Green
    "\001YELLOWm": "\033[38;5;228m",  # Yellow
    "\001BLUEm": "\033[0;34m",  # Blue
    "\001PURPLEm": "\033[38;2;189;147;249m",  # Purple
    "\001CYANm": "\033[38;2;139;233;253m",  # Cyan
    "\001WHITEm": "\033[0;37m",  # White
    "\001PINKm": "\033[38;2;255;121;198m",  # pink
    # Bold
    "\001BOLD_BLACKm": "\033[1;30m",  # Black
    "\001BOLD_REDm": "\033[1;31m",  # Red
    "\001BOLD_GREENm": "\033[1;32m",  # Green
    "\001BOLD_YELLOWm": "\033[1;33m",  # Yellow
    "\001BOLD_BLUEm": "\033[1;34m",  # Blue
    "\001BOLD_PURPLEm": "\033[1;35m",  # Purple
    "\001BOLD_CYANm": "\033[1;36m",  # Cyan
    "\001BOLD_WHITEm": "\033[1;37m",  # White
}


def colorizer(record, can_colorize=True):
    record = str(record)
    record = record.replace(r"\u0001", "\x01")
    if can_colorize:
        for k, v in COLORS.items():
            record = record.replace(k, v)
    else:
        for k, v in COLORS.items():  # pragma: no cover
            record = record.replace(k, "")

    return record


def html_table(dictset, limit: int = 5):  # pragma: no cover
    """
    Render the dictset as a HTML table.

    Parameters:
        dictset: iterable of dictionaries
            The dictset to render
        limit: integer (optional)
            The maximum number of record to show in the table, defaults to 5

    Returns:
        string (HTML table)
    """

    def sanitize(htmlstring):
        ## some types need converting to a string first
        if isinstance(htmlstring, (list, tuple, set)) or hasattr(htmlstring, "as_list"):
            return "[ " + ", ".join([sanitize(i) for i in htmlstring]) + " ]"
        if hasattr(htmlstring, "items"):
            return sanitize("{ " + ", ".join([f'"{k}": {v}' for k, v in htmlstring.items()]) + " }")
        if not isinstance(htmlstring, str):
            return str(htmlstring)
        escapes = {'"': "&quot;", "'": "&#39;", "<": "&lt;", ">": "&gt;", "$": "&#x24;"}
        # This is done first to prevent escaping other escapes.
        htmlstring = htmlstring.replace("&", "&amp;")
        for seq, esc in escapes.items():
            htmlstring = htmlstring.replace(seq, esc)
        return htmlstring

    def _to_html_table(data, columns):
        yield '<table class="table table-sm">'
        for counter, record in enumerate(data):
            if counter == 0:
                yield '<thead class="thead-light"><tr>'
                yield "<td></td>"
                for column in columns:
                    yield f"<th>{sanitize(column)}<th>\n"
                yield "</tr></thead><tbody>"

            yield "<tr>"
            yield f"<td><bold>{counter}</bold></td>"
            for i, column in enumerate(columns):
                sanitized = sanitize(record[i])
                yield f"<td title='{sanitized}' style='max-width:320px;overflow:hidden;text-overflow:ellipsis;white-space:nowrap;'>{sanitized}<td>\n"
            yield "</tr>"

        yield "</tbody></table>"

    footer = f"\n<p>{dictset.rowcount} rows x {dictset.columncount} columns</p>"  # type:ignore

    return "".join(_to_html_table(dictset.head(limit), dictset.column_names)) + footer


def truncate_display(value: str, limit: int = 32) -> str:
    """Truncate the string value for display.

    Parameters:
        value: str
            The value to be truncated.
        limit: int
            The maximum number of characters to display.

    Returns:
        A truncated representation of the value.
    """
    if len(value) > limit:
        return f"{value[:limit - 3]}\033[38;5;102m...\001OFFm"
    return value


def ascii_table(
    table,
    limit: int = 5,
    display_width: Union[bool, int] = True,
    max_column_width: int = 30,
    colorize: bool = True,
    top_and_tail: bool = True,
    show_types: bool = False,
):
    """
    Render the dictset as a ASCII table.

    NOTE:
        This exhausts generators so is only recommended to be used on lists.

    Parameters:
        dictset: iterable of dictionaries
            The dictset to render
        limit: integer (optional)
            The maximum number of record to show in the table, defaults to 5
        display_width: integer/boolean (optional)
            The maximum width of the table, if an integer, the number of characters,
            if a boolean, True uses the display width, False disables (5000)

    Returns:
        string (ASCII table)
    """
    from math import isnan

    import numpy

    if len(table) == 0:
        return "No data in table"

    # get the width of the display
    if isinstance(display_width, bool):
        if not display_width:  # pragma: no cover
            display_width = 5000
        else:
            import shutil

            display_width = shutil.get_terminal_size((80, 20))[0]
    # Extract head data
    if limit > 0 and not top_and_tail:
        t = table.slice(length=limit)
    elif limit > 0 and top_and_tail and table.rowcount > ((2 * limit) + 1):
        t = table.head(size=limit) + table.tail(size=limit)
    else:
        t = table

    # width of index column
    index_width = len(str(len(table))) + 2

    def numpy_type_mapper(value):
        if numpy.issubdtype(value.dtype, numpy.integer):
            return int(value)
        elif numpy.issubdtype(value.dtype, numpy.floating):
            return float(value)
        elif numpy.issubdtype(value.dtype, numpy.bool_):
            return bool(value)
        elif numpy.issubdtype(value.dtype, numpy.ndarray):
            return list(value)
        else:
            return str(value)

    def type_formatter(value, width):
        if isinstance(value, (numpy.generic, numpy.ndarray)):
            value = numpy_type_mapper(value)

        if value is None or (isinstance(value, float) and isnan(value)):
            return "\001NULLm" + truncate_display("null".rjust(width), width) + "\001OFFm"
        if isinstance(value, bool):
            # bool is a superclass of int, do before the int test
            return "\001CONSTm" + truncate_display(str(value).rjust(width), width) + "\001OFFm"
        if isinstance(value, int):
            return "\001INTEGERm" + truncate_display(str(value).rjust(width), width) + "\001OFFm"
        if isinstance(value, (float, decimal.Decimal)):
            return "\001FLOATm" + truncate_display(str(value).rjust(width), width) + "\001OFFm"
        if isinstance(value, str):
            return "\001VARCHARm" + trunc_printable(str(value).ljust(width), width) + "\001OFFm"
        if isinstance(value, datetime.datetime):
            value = f"{value.strftime('%Y-%m-%d')} \001TIMEm{value.strftime('%H:%M:%S')}"
            return "\001DATEm" + (value.rjust(width), width) + "\001OFFm"
        if isinstance(value, datetime.date):
            value = f"{value.strftime('%Y-%m-%d')}"
            return "\001DATEm" + truncate_display(value.rjust(width), width) + "\001OFFm"
        if isinstance(value, (bytes, bytearray)):
            return "\001BLOBm" + truncate_display(str(value).ljust(width), width) + "\001OFFm"
        if isinstance(value, dict):
            value = (
                "\001PUNCm{"
                + "\001PUNCm, ".join(
                    f"'\001KEYm{k}\001PUNCm':'\001VALUEm{v}\001PUNCm'" for k, v in value.items()
                )
                + "}\001OFFm"
            )
            return trunc_printable(value, width)
        if hasattr(value, "days"):
            # MonthDayNano is a superclass of list, do before list
            if isinstance(value, datetime.timedelta):
                days = value.days
                months = 0
                seconds = value.microseconds / 1e6 + value.seconds
            else:
                days = value.days
                months = value.months
                seconds = value.nanoseconds / 1e9

            hours, seconds = divmod(seconds, 3600)
            minutes, seconds = divmod(seconds, 60)
            years, months = divmod(months, 12)
            parts = []
            if years >= 1:
                parts.append(f"{int(years)}y")
            if months >= 1:
                parts.append(f"{int(months)}mo")
            if days >= 1:
                parts.append(f"{int(days)}d")
            if hours >= 1:
                parts.append(f"{int(hours)}h")
            if minutes >= 1:
                parts.append(f"{int(minutes)}m")
            if seconds > 0:
                parts.append(f"{seconds:.2f}s")
            value = f"\001INTERVALm{' '.join(parts)}\001OFFm"
            return trunc_printable(value, width)
        if isinstance(value, (list, tuple)):
            value = (
                "\001PUNCm['\001VALUEm"
                + "\001PUNCm', '\001VALUEm".join(map(str, value))
                + "\001PUNCm']\001OFFm"
            )
            return trunc_printable(value, width)
        return str(value).ljust(width)[:width]

    def character_width(symbol):
        import unicodedata

        category = unicodedata.category(symbol)

        # Get East Asian Width property
        eaw = unicodedata.east_asian_width(symbol)
        # Check for full-width or wide characters
        if eaw in ("F", "W"):
            return 2

        # Check for emojis and certain symbols that take up two spaces in many terminals
        if category == "So" and ord(symbol) > 0x1F000:
            return 2

        # Default case
        return 1

    def trunc_printable(value: str, width: int, full_line: bool = True) -> str:
        """
        Truncates a string to fit within a given display width.

        Parameters:
            value: str
                The string to be truncated.
            width: int
                The maximum display width.
            full_line: bool, optional
                Whether to pad the string with spaces to fill the whole line, defaults to True.

        Returns:
            str: The truncated string.
        """
        offset = 0
        emit = ""
        ignoring = False

        for i, char in enumerate(value):
            if char == "\n":
                emit += "\001CRLFm↵\001VARCHARm"
                offset += 1
                continue
            if char == "\r":
                continue

            emit += char

            if char in ("\033", "\001"):
                ignoring = True

            if not ignoring:
                offset += character_width(char)  # Make sure this function works as expected
            if ignoring and char == "m":
                ignoring = False

            if not ignoring and offset == width:
                break

        line = emit + "\001OFFm"

        if full_line:
            return line + " " * (width - offset)

        return line

    def _inner():
        # Calculate width
        col_width = list(map(len, t.column_names))
        data_width = [
            max(list(map(len, map(str, [p for p in h if p is not None]))) + [4])
            for h in (t.collect(i) for i in range(t.columncount))
        ]
        col_types = [str(t.description[c][1]).lower() for c in range(len(t.column_names))]
        if show_types:
            col_type_width = list(map(len, col_types))
        else:
            col_type_width = [0] * len(col_types)
        col_width = [
            min(max(cw, ctw, dw), max_column_width)
            for cw, ctw, dw in zip(col_width, col_type_width, data_width)
        ]

        # Print data
        yield ("┌" + ("─" * index_width) + "┬─" + "─┬─".join("─" * cw for cw in col_width) + "─┐")
        yield (
            "│"
            + (" " * index_width)
            + "│ "
            + " │ ".join(
                "\001HEADm" + v.center(w)[:w] + "\001OFFm"
                for v, w in zip(t.column_names, col_width)
            )
            + " │"
        )
        if show_types:
            yield (
                "│"
                + (" " * index_width)
                + "│ "
                + " │ ".join(
                    "\001TYPEm" + v.center(w)[:w] + "\001OFFm" for v, w in zip(col_types, col_width)
                )
                + " │"
            )
        yield ("╞" + ("═" * index_width) + "╪═" + "═╪═".join("═" * cw for cw in col_width) + "═╡")
        for i, row in enumerate(t):
            if top_and_tail and (len(t) + 1) < len(table):
                if i == limit:
                    yield "\001PUNCm...\001OFFm"
                if i >= limit:
                    i += len(table) - (limit * 2)
            formatted = [type_formatter(v, w) for v, w in zip(row, col_width)]
            yield (
                "│\001TYPEm"
                + str(i + 1).rjust(index_width - 1)
                + "\001OFFm │ "
                + " │ ".join(formatted)
                + " │"
            )
        yield ("└" + ("─" * index_width) + "┴─" + "─┴─".join("─" * cw for cw in col_width) + "─┘")

    return "\n".join(
        colorizer(trunc_printable(line, display_width, False), colorize) for line in _inner()
    )


def markdown(
    table,
    limit: int = 5,
    max_column_width: int = 30,
):
    # Extract head data
    if limit > 0:
        t = table.slice(length=limit)
    else:
        t = table

    # width of index column
    index_width = len(str(len(table)))

    # Calculate width
    col_width = list(map(len, t.column_names))
    data_width = [
        max(list(map(len, map(str, [p for p in h if p is not None]))) + [4])
        for h in (t.collect(i) for i in range(t.columncount))
    ]
    col_width = [min(max(cw, dw), max_column_width) for cw, dw in zip(col_width, data_width)]

    # Print data
    data = [t.row(i) for i in range(len(t))]
    yield (
        "| #"
        + (" " * (index_width - 2))
        + "| "
        + " | ".join(v.ljust(w)[:w] for v, w in zip(t.column_names, col_width))
        + " |"
    )
    yield ("|" + ("-" * index_width) + "|-" + "-|-".join("-" * cw for cw in col_width) + "-|")
    for i in range(len(data)):
        formatted = [str(v).rjust(w)[:w] for v, w in zip(data[i], col_width)]
        yield ("|" + str(i + 1).rjust(index_width - 1) + " | " + " | ".join(formatted) + " |")
